import numpy as np
from mne import Epochs, EpochsArray

from .helpers import round_like_matlab


def correct_trials(results, data, rt=None):
    """Subtracts the 'r' component from the single trial data, shifted by the RTs."""

    if isinstance(data, (Epochs, EpochsArray)):
        is_epochs = True
        epochs = data.copy()
        ch_types = np.array(epochs.get_channel_types())
        eeg_ixs = np.where(ch_types == 'eeg')[0]
        data = epochs.get_data(picks='eeg').copy()
        data = np.swapaxes(data, 0, 2)

    n_trials = data.shape[2]

    if rt is None:
        rt = results.latencies['r'].copy()
        n_rt = len(rt)
        assert n_trials == n_rt, \
            'When passing different single trial data than the ones used ' + \
            'to perform RIDE, you must also pass the corresponding RTs'
    else:
        rt = rt / results.cfg.re_samp
        rt = round_like_matlab(rt - np.median(rt))
        rt = rt * results.cfg.re_samp

    srate = 1000.0 / results.cfg.samp_interval

    data_corr = data - move3(np.repeat(results.comps['r'][:, :, np.newaxis],
                                       n_trials,
                                       axis=2),
                             np.round(rt*srate/1000).astype(int).flatten())

    if not is_epochs:

        return data_corr

    data_corr = np.swapaxes(data_corr, 0, 2)
    epochs._data[:, eeg_ixs, :] = data_corr

    return epochs


def move3(data, latency):
    """Shifts single trial data ('r' component) by the single trial latency."""

    latency = np.round(latency).astype(int)
    d1, d2, d3 = data.shape
    temp = np.zeros((d1, d2, d3), dtype=data.dtype)

    left = latency + 1
    left[left <= 0] = 1
    right = d1 + latency
    right[right > d1] = d1

    left1 = -latency + 1
    left1[left1 <= 0] = 1
    right1 = d1 - latency
    right1[right1 > d1] = d1

    for j in np.where((latency > -d1) & (latency <= d1))[0]:
        temp[left[j] - 1:right[j], :, j] = data[left1[j] - 1:right1[j], :, j]

    return temp



import numpy as np


def round_like_matlab(x):
    """Round to nearest integer, like MATLAB's round() function."""

    return np.trunc(x + np.copysign(0.5, x)).astype(int)


def mean_nan(data, dim):

    data = np.nan_to_num(data)
    f = np.mean(data, dim)

    return f



import numpy as np


def ride_detrend(data, twd):

    # # For debugging
    # data = temp.copy()
    # twd = np.array([51, 110, 288, 348])

    index = np.isnan(data)
    data[index] = 0.0
    d1, d2 = data.shape
    d3 = 1  # MATLAB uses `d1, d2, d3 = data.shape` even though `data` is 2D
    d = [(twd[3] + twd[2]) / 2 - (twd[1] + twd[0]) / 2]

    temp0 = np.tile(np.arange(d1) + 1, [d2, d3]).T

    a = data[np.arange(twd[0], twd[1] + 1), :].mean(axis=0, keepdims=True)
    b = data[np.arange(twd[2], twd[3] + 1), :].mean(axis=0, keepdims=True)
    temp = (b - a) / d
    data = data - temp0 * temp[np.zeros(d1, dtype=int), :]

    temp = data[np.arange(twd[0], twd[1] + 1), :].mean(axis=0, keepdims=True)
    data = data - temp[np.zeros(d1, dtype=int), :]

    f = data.copy()
    f[index] = np.nan

    return f



import os
import sys

import matplotlib.pyplot as plt
import numpy as np
from pyprojroot.here import here
from scipy.io import loadmat

sys.path.insert(1, os.path.join(sys.path[0], '..'))

from ride.cfg import RideCfg
from ride.ride import ride_call

mat_file = here('matlab/matlab_example/mat/Vp0001_rep1_close.mat')
data_dict = loadmat(str(mat_file), mat_dtype=True)
data = data_dict['data']
data = data.astype('float64')
rt = data_dict['rt']
rt = rt.astype('float64')


cfg = RideCfg(comp_name=['s', 'r'],
              comp_twd=[[0, 600], [-300, 300]],
              comp_latency=[0, rt],
              sfreq=500)

results = ride_call(data, cfg)

n_comps = len(results.comps)
ncols = 1 + n_comps
fig, axs = plt.subplots(1, ncols, figsize=(10, 3))
x = np.linspace(cfg.epoch_twd[0], cfg.epoch_twd[1], data.shape[0])

axs[0].plot(x, results.erp, linewidth=0.7)
axs[0].set_title('erp')
axs[0].set_ylim([-20.0, 20.0])
axs[0].set_xlabel('Time (ms)')
axs[0].set_ylabel('Amplitude (µV)')

for ax, key in zip(axs[1:], results.comps):
    ax.plot(x, results.comps[key], linewidth=0.7)
    ax.set_title(key)
    ax.set_ylim([-20.0, 20.0])
    ax.set_xlabel('Time (ms)')

fig.show()



import numpy as np
from scipy.interpolate import CubicSpline


def interp2d(data, x1, x2):

    temp = np.empty((len(x2), data.shape[1]))

    for j in np.arange(data.shape[1]):
        cs = CubicSpline(x1, data[:, j], axis=0)
        temp[:, j] = cs(x2)

    return temp



from warnings import warn

import numpy as np
from mne import Epochs, EpochsArray

from .baseline import baseline
from .helpers import round_like_matlab
from .interp2d import interp2d
from .iter import ride_iter
from .results import RideResults

def ride_call(data, cfg):

    # # Load example data
    # from pyprojroot.here import here
    # mat_file = here('matlab/matlab_example/mat/Vp0008_rep1_distant.mat')
    # data_dict = loadmat(str(mat_file), mat_dtype=True)
    # data = data_dict['data']
    # data = data.astype('float64')
    # rt = data_dict['rt']
    # rt = rt.astype('float64')

    # start RIDE correction
    assert len(cfg.comp_name) > 1, 'At least two components are required'

    # TODO: make sure order is always s, (c), r

    cfg0 = cfg.copy()

    # If necessary, extract Numpy arra from MNE epochs object
    if isinstance(data, (Epochs, EpochsArray)):
        data = data.get_data(picks='eeg')
        data = np.swapaxes(data, 0, 2)

    # section 1
    d1, d2, d3 = data.shape
    epoch_length = d1
    erp = data.mean(axis=2)
    results = RideResults(erp=erp, latency0=cfg.comp_latency.copy())

    rs = cfg.re_samp / cfg.samp_interval
    data = data[round_like_matlab(np.linspace(0, d1 - 1, int(d1 / rs))), :, :]
    d1, d2, d3 = data.shape  # New size after down samping

    for j in range(cfg.comp_num):

        if isinstance(cfg.comp_latency[j], (int, float)):
            int_value = cfg.comp_latency[j]
            if cfg.prg == 1:
                print(f'WARNING: Extending integer latency {int_value} to a vector of {int_value}s (one per trial)')
            cfg.comp_latency[j] = np.array([[int_value]] * d3)

        if cfg.comp_name[j] == 'r':
            cfg.comp_twd_samp[j] = cfg.comp_twd_samp[j] + np.median(results.latency0[j])
            cfg.comp_twd_samp[j][cfg.comp_twd_samp[j] < cfg.rwd] = cfg.rwd
            cfg.comp_twd_samp[j][cfg.comp_twd_samp[j] > cfg.epoch_twd[1]] = cfg.epoch_twd[1]

        cfg.comp_latency[j] = cfg.comp_latency[j] / cfg.re_samp
        cfg.comp_latency[j] = round_like_matlab(cfg.comp_latency[j]-np.median(cfg.comp_latency[j])) 

        cfg.comp_twd_samp[j] = np.fix((cfg.comp_twd_samp[j] - cfg.epoch_twd[0])/cfg.re_samp)+[1, -1]

    stop = 1

    # TODO: Add "outer iteration" loop around here if there are one or more C components

    cfg1 = cfg.copy()
    cfg1.final = stop
    cfg1.inner_iter = 100

    c_l = np.zeros((d1, cfg.comp_num, d2))
    c_sl = c_l.copy()

    amp = np.zeros((d3, d2, cfg.comp_num))
    for c in range(d2):

        if cfg.prg == 1:
            print(f'Processing electrode #{c + 1}')

        rst = ride_iter(np.squeeze(data[:, c, :]), cfg1)

        c_l[:, :, c] = rst['comp']
        c_sl[:, :, c] = rst['comp1']

        if stop == 1:
            amp[:, c, :] = rst['amp']
        
        if cfg.prg == 1:
            print(f'Took {rst["iter"] + 1} iterations')

    comp = c_l.transpose((0, 2, 1))
    comp1 = c_sl.transpose((0, 2, 1))

    results.erp_new = 0
    results.residue = erp

    bl_wd = np.arange(-cfg.epoch_twd[0]/cfg.samp_interval,
                    -cfg.epoch_twd[0]/cfg.samp_interval+cfg.bl/cfg.samp_interval,
                    dtype=int)

    component = np.zeros((d1, d2, cfg.comp_num))
    component1 = np.zeros((d1, d2, cfg.comp_num))

    for j in np.arange(cfg.comp_num):
        # The MATLAB version explicitly requests "spline" as the interpolation method
        # Our Python function `interp2d` only performs this spline interpolation,
        # with no support for any other method (unlike the MATLAB function)
        component[:, : , j] = interp2d(comp[:, :, j],
                                    np.round(np.linspace(0, epoch_length, d1, endpoint=False)),
                                    np.arange(0, epoch_length))
        component[:, :, j] = baseline(component[:, :, j],bl_wd)
        component1[:, :, j] = interp2d(comp1[:, :, j],
                                        np.round(np.linspace(0, epoch_length, d1, endpoint=False)),
                                        np.arange(0, epoch_length))
        component1[:, :, j] = baseline(component1[:, :, j],bl_wd)
        results.residue = results.residue - component1[:, :, j]
        results.comps[cfg.comp_name[j]] = component[:, :, j]
        results.comps_sl[cfg.comp_name[j]] = component1[:, :, j]
        results.latencies[cfg.comp_name[j]] = cfg.comp_latency[j]*cfg.re_samp
        results.amps[cfg.comp_name[j]] = amp[:, :, j]

    results.comps[cfg.comp_name[0]] = baseline(results.comps[cfg.comp_name[0]],bl_wd) + np.repeat(np.mean(erp[bl_wd, :], axis=0)[np.newaxis, :], epoch_length, axis=0)
    results.comps_sl[cfg.comp_name[0]] = baseline(results.comps_sl[cfg.comp_name[0]],bl_wd) + np.repeat(np.mean(erp[bl_wd, :], axis=0)[np.newaxis, :], epoch_length, axis=0)

    results.comps[cfg.comp_name[rst['trend_c']]] = baseline(results.comps[cfg.comp_name[rst['trend_c']]] + results.residue,bl_wd)
    results.comps_sl[cfg.comp_name[rst['trend_c']]] = baseline(results.comps_sl[cfg.comp_name[rst['trend_c']]] + results.residue,bl_wd)

    if  cfg.comp_num == 1:
        bl_wd = np.arange(-cfg.epoch_twd[0]/cfg.samp_interval, dtype='int')
        results.comps[cfg.comp_name[0]] = baseline(results.comps[cfg.comp_name[0]] + results.residue,bl_wd) + np.repeat(np.mean(erp[bl_wd, :], axis=0)[np.newaxis, :], epoch_length, axis=0)
        results.comps_sl[cfg.comp_name[0]] = baseline(results.comps_sl[cfg.comp_name[0]] + results.residue,bl_wd) + np.repeat(np.mean(erp[bl_wd, :], axis=0)[np.newaxis, :], epoch_length, axis=0)

    for j in np.arange(cfg.comp_num):
        results.erp_new = results.erp_new + results.comps[cfg.comp_name[j]]

    results.cfg = cfg0.copy()

    return results



import numpy as np


def baseline(x, a=None):

    x[np.isnan(x)] = 0.0

    if a is None:
        temp = np.mean(x, axis=0)
        f = x - temp[np.newaxis, :]
    else:
        temp = np.mean(x[a, :], axis=0)
        f = x - temp[np.newaxis, :]

    return f



import numpy as np

from .helpers import round_like_matlab


def median_2d(x):

    # !! Median is probably biased if nan trials are present b/c they are set to zero
    # TODO: simulate this and see if it's a problem
    x = x.copy()
    x[np.isnan(x)] = 0.0  # zeros padding
    s = x.shape
    x = np.sort(x, axis=0)
    y = x.flatten(order='F')[(round_like_matlab(
        s[0]/2) + np.arange(0, s[1])*s[0])-1]

    return y



import matplotlib.pyplot as plt
import numpy as np


class RideResults:

    def __init__(self,
                 erp,
                 latency0,
                 erp_new=None,
                 residue=None,
                 comps=None,
                 comps_sl=None,
                 latencies=None,
                 amps=None,
                 cfg=None):

        if comps is None:
            comps = {}

        if comps_sl is None:
            comps_sl = {}

        if latencies is None:
            latencies = {}

        if amps is None:
            amps = {}

        self.erp = erp
        self.latency0 = latency0
        self.erp_new = erp_new
        self.residue = residue
        self.comps = comps
        self.comps_sl = comps_sl
        self.latencies = latencies
        self.amps = amps
        self.cfg = cfg

    def plot(self):
        """Plots time courses of the ERP and RIDE components."""

        n_comps = len(self.comps)
        ncols = 1 + n_comps
        fig, axs = plt.subplots(1, ncols, figsize=(10, 3))

        x = np.linspace(self.cfg.epoch_twd[0],
                        self.cfg.epoch_twd[1],
                        self.erp.shape[0])

        y_lim = self._get_ylim()

        axs[0].plot(x, self.erp, linewidth=0.7)
        axs[0].set_title('erp')
        axs[0].set_xlabel('Time')
        axs[0].set_ylabel('Amplitude')
        axs[0].set_ylim(y_lim)

        for ax, key in zip(axs[1:], self.comps):
            ax.plot(x, self.comps[key], linewidth=0.7)
            ax.set_title(key)
            ax.set_ylim(y_lim)
            ax.set_xlabel('Time')

        return fig

    def _get_ylim(self):
        """Gets appropriate y-axis limits for the time course plot."""

        y_max_erp = np.abs(self.erp.max())
        y_max_comps = [np.abs(comp.max()) for comp in self.comps.values()]
        y_max = max(y_max_erp, *y_max_comps)

        # Round up "small" numbers (e.g., volts) to nearest decimal
        if y_max < 1.0:
            n_digits = int(np.ceil(-np.log10(np.abs(y_max))))
            y_max = np.ceil(y_max * 10 ** n_digits) / 10 ** n_digits
        else:  # Round up "large" numbers (e.g., microvolts) to nearest integer
            y_max = np.ceil(y_max)

        return [-y_max, y_max]



import numpy as np


def ride_tukey(n, r):
    n = int(n)
    t = np.linspace(0, 1, n)
    # Define period of the taper as 1/2 period of a sine wave.
    per = r/2
    tl = int(np.floor(per*(n-1))+1)
    th = n-tl+1
    # Window is defined in three sections: taper, constant, taper
    f = np.concatenate([(1+np.cos(np.pi/per*(t[np.arange(tl)] - per)))/2,
                       np.ones((th-tl-1)),
                       (1+np.cos(np.pi/per*(t[th-1:] - 1 + per)))/2])

    if n == 1:
        f = 1

    return f



import matplotlib.pyplot as plt
import numpy as np


def generate_sine_data(
        s_duration=500.0, s_amplitude=15.0, r_duration=200.0, r_amplitude=5.0,
        r_latencies=sorted(np.random.normal(250.0, 50.0, size=100))):
    """
    Generates sine wave data.

    The S component is one long bump (half sine wave) with a fixed onset
    latency of zero. R is a short bump (half sine wave) riding on top of S
    with a variable latency.

    See Figure 1 in Ouyang et al. (2015), https://doi.org/10.1016/j.jneumeth.2014.10.009

    Parameters
    ----------
    s_duration : float
        The duration (in ms) of the S component. This will be the same as the
        duration (first dimension) of the generated data.
    s_amplitude : float
        The amplitude of the S component in (in microvolts).
    r_duration : float
        The duration (in ms) of the R component. Should be shorter than
        `s_duration`.
    r_amplitude : float
        The amplitude of the S component in (in microvolts).
    r_latencies : iterable
        The single trial (peak) latencies of the R component. The length
        of this object will determine the number of trials (third dimension)
        of the generated data.

    Returns
    -------
    data : ndarray (n_times, n_channels, n_trials)
        The generated data (sum of S and single-trial R components). The
        number of channels (second dimension) will always be one.
    """

    times_s = np.arange(s_duration)
    s = s_amplitude * np.sin(np.pi * times_s / s_duration)

    times_r = np.arange(r_duration)
    r = r_amplitude * np.sin(np.pi * times_r / r_duration)

    n_times = int(s_duration)
    n_channels = 1
    n_trials = len(r_latencies)
    data = np.zeros((n_times, n_channels, n_trials))

    channel_ix = 0
    for trial_ix, r_latency in enumerate(r_latencies):
        data[:, channel_ix, trial_ix] = s
        r_start_ix = int(r_latency - r_duration / 2.0)
        r_end_ix = int(r_latency + r_duration / 2.0)
        data[r_start_ix:r_end_ix, channel_ix, trial_ix] += r

    return data, r_latencies


def plot_input_data(data):
    """
    Plots single trial and averaged input data before RIDE.

    TODO: Create one subplot for each channel instead of just plotting the
    first channel.

    Parameters
    ----------
    data : ndarray (n_times, n_channels, n_trials)
        The single trial input data.
    """

    for trial_ix in range(data.shape[2]):
        plt.plot(data[:, 0, trial_ix], color='black', alpha=0.1)

    erp = data.mean(axis=2)
    plt.plot(erp[:, 0], linewidth=4.0, color='red')

    plt.xlabel('Time (ms)')
    plt.ylabel('Amplitude (µV)')



import numpy as np
from scipy.fft import fft, ifft


def filtering20(x, a, b):

    f = x
    for j in range(x.shape[1]):
        f[:, j] = filtering10(x[:, j], a, b)

    return f


def filtering10(x, a, b):

    # The MATLAB code does x = x(:) here but this doesn't seem to change anything
    x0 = x.mean()
    x = x - x0
    n = 10 * len(x)
    Y = fft(x, n)

    b = int(b)
    H = np.concatenate([np.zeros((1, a)),
                        np.expand_dims(Y[a: b + 1], axis=0),
                        np.zeros((1, n - b * 2 - 1)),
                        np.expand_dims(Y[n - b: n - a + 1], axis=0),
                        np.zeros((1, a - 1))], axis=1)
    H = H.transpose()

    f = ifft(H, axis=0).real
    f = np.squeeze(f)

    return f[:len(x)]



import numpy as np

from .detrend import ride_detrend
from .filtering import filtering20
from .helpers import mean_nan
from .median2d import median_2d
from .tukey import ride_tukey


def ride_iter(data, cfg):

    # # For debugging
    # cfg = cfg1
    # data = data[:, 61, :]

    d1, d2 = data.shape
    bd = cfg.bd

    data0 = data.copy()

    data = filtering20(data, 1, np.fix(10 * 20 * cfg.re_samp * d1 / 1000))
    # TODO: Compare with MNE filter

    max_latency = np.zeros((cfg.comp_num), dtype=int)
    min_latency = np.zeros((cfg.comp_num), dtype=int)
    length_c = np.zeros((cfg.comp_num), dtype=int)
    for j in range(cfg.comp_num):
        max_latency[j] = cfg.comp_latency[j].max()
        min_latency[j] = cfg.comp_latency[j].min()
        length_c[j] = d1 + max_latency[j] - min_latency[j]

    com_c = np.zeros((d1, cfg.comp_num))
    com_c1 = np.zeros((d1, d2, cfg.comp_num))
    amp_c = np.zeros((d2, cfg.comp_num))

    trend_i = cfg.comp_num-1
    stream_flow = np.arange(cfg.comp_num)
    for c in range(cfg.comp_num):
        if cfg.comp_name[c] == 'r':
            trend_i = c-1
    if trend_i == cfg.comp_num:
        stream_flow = np.concatenate([np.array([trend_i]), np.arange(trend_i)])
    if trend_i == cfg.comp_num-2 : 
        stream_flow = np.concatenate([np.array([trend_i]), np.arange(trend_i), np.array([cfg.comp_num-1])])
    if trend_i == 0 : 
        stream_flow = np.array([1,0])

    l1 = np.zeros((cfg.inner_iter, cfg.comp_num))
    stop = 0
    stop_c = np.zeros((cfg.comp_num, 1))
    com_old = com_c.copy()

    # for iter in np.arange(1, 2): # For testing purposes, only runs iter = 1
    for iter in np.arange(cfg.inner_iter):

        if iter + 1 == cfg.inner_iter:
            stop = 1

        # track the convergence

        # !!!! CONTINUE HERE, AFTER FINISHING FIRST ITERATION OF FOR LOOP
        # TO DO: add if-loop for iter > 1

        if iter > 1: 
            for c in np.arange(cfg.comp_num):
                l1[iter-2,c]=np.sum(np.abs(com_c[:,c] - com_old[:,c]))
                if iter > 2:
                    if l1[iter-3,c] - l1[iter-2,c] < 0.01*(l1[0,c] - l1[1,c]): # convergence is now defined as 0.01
                        stop_c[c] = 1
                    if l1[iter-3,c] == l1[iter-2,c]:
                        stop_c[c] = 1
            if cfg.comp_num == 1:
                stop_c[c] = 1
        for c in np.arange(cfg.comp_num):
            com_old[:,c] = com_c[:,c]
            # decision of the termination of the interation of each component
            

        
        #for c in np.arange(cfg.comp_num):
            #com_old[:,c]=com_c[:,c] # decision of the termination of iteration of each component
        ## CAUTION: for-loop seems unnecessary

        # com_old = com_c.copy()

        for c in stream_flow:
            if stop_c[c] == 0:
                temp = data.copy()
                for j in np.arange(cfg.comp_num):
                    if j != c:
                        temp = temp-com_c1[:, :, j]
                residue = temp.copy()
                temp = np.empty((length_c[c], d2))
                temp[:] = np.nan
                for j in np.arange(d2):
                    temp[np.arange(-cfg.comp_latency[c][j]+max_latency[c],d1-cfg.comp_latency[c][j]+max_latency[c]),j] = residue[:,j]

                temp = ride_detrend(
                    temp,
                    np.array([0,
                              np.fix((cfg.comp_twd_samp[c][1] - cfg.comp_twd_samp[c][0]) * bd),
                              np.fix((cfg.comp_twd_samp[c][1] - cfg.comp_twd_samp[c][0]) * (1 - bd)) - 1,
                              cfg.comp_twd_samp[c][1] - cfg.comp_twd_samp[c][0] - 1], dtype=int) \
                                + max_latency[c] + int(cfg.comp_twd_samp[c][0]))
                
                temp0 = median_2d(temp.T)
                temp0[np.concatenate([np.arange(cfg.comp_twd_samp[c][0], dtype=int) + max_latency[c],
                                      np.arange(cfg.comp_twd_samp[c][1] + max_latency[c]-1, length_c[c], dtype=int)])]=0
                temp0[np.arange(cfg.comp_twd_samp[c][0]+ max_latency[c]-1, cfg.comp_twd_samp[c][1] + max_latency[c], dtype=int)] = \
                    temp0[np.arange(cfg.comp_twd_samp[c][0]+ max_latency[c]-1, cfg.comp_twd_samp[c][1] + max_latency[c], 
                                    dtype=int)] * ride_tukey(cfg.comp_twd_samp[c][1] - cfg.comp_twd_samp[c][0] + 1, bd*2)
     
                temp1 = np.repeat(temp0[:,np.newaxis],d2, axis=1)
                temp1[np.isnan(temp)] = np.nan

                # TODO: !!!! IMPORTANT: Next line is only for testing purposes, delete later!!
                # temp1[:,44] = np.nan     # # TODO: !!!! IMPORTANT: This line is only for testing purposes, delete later!!
                temp1 = np.reshape(temp1.T[~np.isnan(temp1.T)], (d1, d2), order='F')
                com_c[:,c] = temp0[np.arange(max_latency[c], max_latency[c]+d1)]
                com_c1[:,:,c] = temp1
            # End of if-loop
        # End of stream-flow for loop

        if stop_c.mean() == 1:
            stop = 1
        if stop == 1:
            for c in stream_flow:

                temp = data.copy()
                for j in np.arange(cfg.comp_num):
                    if j != c:
                        temp = temp - com_c1[:,:,j]

                residue = temp.copy()
                temp = np.empty((length_c[c], d2))
                temp[:] = np.nan

                for j in np.arange(d2):
                    temp[np.arange(-cfg.comp_latency[c][j]+max_latency[c],
                                   d1-cfg.comp_latency[c][j]+max_latency[c]),
                         j] = residue[:,j]

                if cfg.final == 1:
                    temp[np.isnan(temp)] = 0
                    amp_c[:,c] = np.mean(
                        np.repeat(com_c[np.arange(cfg.comp_twd_samp[c][0] - 1, cfg.comp_twd_samp[c][1], dtype=int), c, np.newaxis], d2, axis=1) * temp[max_latency[c] + np.arange(cfg.comp_twd_samp[c][0] - 1, cfg.comp_twd_samp[c][1], dtype=int), :], axis=0)

            break
        # end of inner iter loop
    
    # for last iteration
    if cfg.final == 1:
        # release time window function and detrending
        # allocate trend to the last C component
        for c in stream_flow:
            temp = data0.copy()
            for j in np.arange(cfg.comp_num):
                if j != c:
                    temp = temp - com_c1[:,:,j]
                    # TODO: Probably only for microsaccades, temp-com_ms1 if 'ms' and 'var' exist
            residue = temp.copy()
            temp = np.empty((length_c[c], d2))
            temp[:] = np.nan

            for j in np.arange(d2):
                temp[np.arange(-cfg.comp_latency[c][j]+max_latency[c],
                               d1-cfg.comp_latency[c][j]+max_latency[c]),
                               j] = residue[:,j]
            temp0 = mean_nan(temp,1)

            if c==stream_flow[0]:
                temp0[np.arange(cfg.comp_twd_samp[c][0] + max_latency[c],dtype=int)]=0.0
                tem = ride_tukey(cfg.comp_twd_samp[c][1] - cfg.comp_twd_samp[c][0] + 1, bd*2)
                tem = tem[np.arange(np.fix(len(tem)/2).astype(int))]
                temp0[np.arange(cfg.comp_twd_samp[c][0] + max_latency[c]-1, 
                                cfg.comp_twd_samp[c][0] + max_latency[c] + len(tem)-1, dtype=int)] = \
                                    temp0[np.arange(cfg.comp_twd_samp[c][0] + max_latency[c]-1,\
                                                    cfg.comp_twd_samp[c][0] + max_latency[c] + len(tem)-1, dtype=int)] * tem
                
            temp1 = np.repeat(temp0[:,np.newaxis],d2, axis=1)
            temp1[np.isnan(temp)] = np.nan
            temp1 = np.reshape(temp1.T[~np.isnan(temp1.T)], (d1, d2), order='F')
            com_c[:,c] = temp0[np.arange(max_latency[c], max_latency[c]+d1)]
            com_c1[:,:,c] = temp1.copy()
        
        c = stream_flow[0]
        temp = data0.copy()
        for j in np.arange(cfg.comp_num):
            if j != c:
                temp = temp - com_c1[:,:,j]
                # TODO: Probably only for microsaccades, temp-com_ms1 if 'ms' and 'var' exist
        residue=temp.copy()
        temp = np.empty((length_c[c], d2))
        temp[:] = np.nan
        for j in np.arange(d2):
            temp[np.arange(-cfg.comp_latency[c][j]+max_latency[c],
                           d1-cfg.comp_latency[c][j]+max_latency[c]),
                           j] = residue[:,j]
        temp0 = mean_nan(temp,1)
        temp1 = np.repeat(temp0[:,np.newaxis],d2, axis=1)
        temp1[np.isnan(temp)] = np.nan
        temp1 = np.reshape(temp1.T[~np.isnan(temp1.T)], (d1, d2), order='F')
        com_c[:,c] = temp0[np.arange(max_latency[c], max_latency[c]+d1)]
        com_c1[:,:,c] = temp1.copy()

    # end of last iteration
    results={}
    results['amp'] = amp_c
    results['comp'] = com_c
    results['comp1'] = np.mean(com_c1,1)
    results['iter'] = iter
    results['l1'] = l1
    results['trend_c'] = stream_flow[0]

    return results



from setuptools import find_packages, setup

if __name__ == "__main__":

    # Paste README as long description
    with open('README.md', 'r', encoding='utf-8') as fh:
        long_description = fh.read()

    # Actual setup
    setup(
        name='eeg-ride',
        author='Kirsten Stark, Alexander Enge',
        author_email='kirsten.stark@hu-berlin.de, alexander.enge@hu-berlin.de',
        description='Removing speech artifacts from EEG data using Residue '
                    'Iteration Decomposition (RIDE)',
        long_description=long_description,
        long_description_content_type='text/markdown',
        url='https://github.com/kirstenstark/eeg-ride',
        project_urls={
            'Issue trackers': 'https://github.com/kirstenstark/eeg-ride/issues',
        },
        classifiers=[
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Programming Language :: Python :: 3.11',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
        ],
        packages=find_packages(),
        install_requires=[
            'matplotlib',
            'numpy',
            'scipy'
        ],
        python_requires='>=3.8',
    )



from warnings import warn

import numpy as np


class RideCfg:

    def __init__(self,
                 comp_name,
                 comp_twd,
                 comp_latency,
                 sfreq,
                 epoch_twd=None,
                 rwd=200,
                 re_samp=2,
                 bd=0.2,
                 prg=1,
                 bl=200,
                 comp_twd_samp=None):

        assert len(comp_name) == len(comp_twd) == len(comp_latency), \
            '`comp_name`, `comp_twd`, and `comp_latency` must have the same length'

        comp_num = len(comp_name)

        samp_interval = 1000 / sfreq

        if epoch_twd is None:
            epoch_twd = np.array([-100, 1198])
            if epoch_twd[1] < 1000:
                warn('`epoch_twd` is less than 1000 ms post stimulus onset ' +
                     'and therefore probably to short for RIDE correction.')

        if comp_twd_samp is None:
            comp_twd_samp = comp_twd.copy()

        self.comp_name = comp_name
        self.comp_twd = comp_twd
        self.comp_latency = comp_latency
        self.comp_num = comp_num
        self.sfreq = sfreq
        self.samp_interval = samp_interval
        self.epoch_twd = epoch_twd
        self.rwd = rwd
        self.re_samp = re_samp
        self.bd = bd
        self.prg = prg
        self.bl = bl
        self.comp_twd_samp = comp_twd_samp

    def copy(self):

        return RideCfg(comp_name=self.comp_name,
                       comp_twd=self.comp_twd,
                       comp_latency=self.comp_latency,
                       sfreq=self.sfreq,
                       epoch_twd=self.epoch_twd,
                       rwd=self.rwd,
                       re_samp=self.re_samp,
                       bd=self.bd,
                       prg=self.prg,
                       bl=self.bl,
                       comp_twd_samp=self.comp_twd_samp)
